/*****************************************************************************/
/* SAS implementation of supplemented EM algorithm and
/* An asymptotically chi-squared GOF statistic for
/* Covariance structure modelling under multinormal MAR
/* Written by Taehun Lee & Li Cai (Dec 2006 - Aug 2008)
/* L. L. Thurstone Psychometric Lab, UNC
/* and GSE & IS, UCLA
/*****************************************************************************/

%let maxEC=5000;  * maximum number of EM cycles;
%let emest=emest; * where to save the EM estimates;
%let calisoutest=calisoutest; * where to save estimates from Proc Calis;
%let calisoutstat=calisoutstat; * where to save residuals from Proc Calis ;

/* A wrapper macro for Proc Calis, analyzing augmented moment matrix */
%macro FitModel(nobs=,model=);
  proc calis ucov data=&emest augment outjac outest=&calisoutest
    residual outstat=&calisoutstat nobs=&nobs;
    &model;
  run;
  * pull out Jacobian;
  data jacob;
    set &calisoutest (where=(_TYPE_='JACOBIAN'));
    drop _TYPE_ _NAME_ _RHS_ _ITER_ ;
  run;
  * pull out parameter estimates;
  data parmests;
    set &calisoutest (where=(_TYPE_='PARMS'));
	drop _TYPE_ _NAME_ _RHS_ _ITER_ ;
  run;
  * pull out parameter standard errors;
  data wrongstderrs;
    set &calisoutest (where=(_TYPE_='STDERR'));
	drop _TYPE_ _NAME_ _RHS_ _ITER_ ;
  run;
  * pull out UCOV;
  data uncov;
    set &calisoutstat (where=(_TYPE_='UCOV'));
    drop _TYPE_ _NAME_;
  run;
  * pull out MAXPRED;
  data maxpred;
    set &calisoutstat (where=(_TYPE_='MAXPRED'));
    drop _TYPE_ _NAME_;
  run;
%mend FitModel;

/* A collection of various IML subroutines for later use */
proc iml;
/**********************************************************/
/* compute vech(X)   
/* vx-output, x-input, n-number of rows(columns)
/**********************************************************/
start vech(vx,x,n);
  count=0;
  vx=J(n*(n+1)/2,1,0);
  do j=1 to n;
    do i=1 to j;
      count=count+1;
      vx(|count,1|)=x(|i,j|);
    end;
  end;
finish vech;

/**********************************************************/
/* convert veched form into symmetric matrix
/* sm-symmetric output, vm-veched input, 
/* p-number of variables
/**********************************************************/
start symm(sm,vm,p);
  count=0;
  sm=j(p,p,0);
  do j=1 to p;
    do i=1 to j;			
      count=count+1;
      sm[i,j]=vm[count,1];
    end;
  end;
  sm=sm+sm`-diag(sm);
finish symm;

/**********************************************************/
/* convert (mean, vech(covariance matrix) into
/* (-1 || mean` // mean || variance-covariance matrix)
/* rx-rectangular output, x-input vector
/* p: number of variables
/**********************************************************/
start v2r(rx,x,p);
  mean=x[1:p,1];
  v_cov=x[p+1:p*(p+3)/2,1];
  run symm(sm,v_cov,p);
  rx1={-1} || mean`;
  rx2=mean || sm;
  rx=rx1//rx2;
finish v2r;

/**********************************************************/
/* obtain (mean, vech(covariance matrix) from
/* (-1 || mean` // mean || variance-covariance matrix)
/* vx-vector output, x-input matrix, p-number of variables
/**********************************************************/
start r2v(vx, x, p);
  vx1=x[2:p+1,1];
  run vech(vx2, x[2:p+1,2:p+1],p);
  vx=vx1//vx2;
finish r2v;

/**********************************************************/
/* compute duplication matrix, D /  vec(X) = D * vech(X) 
/* order of D is n*n by n*(n+2)/2
/* n-number of rows of symmetric matrix X
/**********************************************************/
start Dup(D,n);
  count = 0;
  D = J(n*n,n*(n+1)/2,0);  
  do col=1 to n;
    do row=1 to col;
      count = count+1;	
      D(|(row-1)*n+col,count|) = 1;
      if col ^= row then 
        D(|(col-1)*n+row,count|) = 1;      		
    end;
  end;
finish Dup; 

/***********************************************************/
/* EM map - one iteration in EM algorithm
/* A-output matrix, theta_0-initial input matrix, 
/* n-sample size, p-number of variable, 
/* R-missing pattern(1:observed,0:missing), Y-data matrix
/***********************************************************/
start EMmap(A, theta_0, n, p, R, Y );
  /* T-matrix for expected sufficient statistics 
     given observed Y and current parameter estimates)*/
  T=j(p+1,p+1,0); 		
  do i=1 to n;		
    A=theta_0; 
    O_s=loc({0}|| R[i,]); 		
    A=sweep(A, O_s);
    A[,O_s]=-A[,O_s]; 									
    do j=1 to p;
      if Y[i,j]=. then Y[i,j]=0;
    end;
	
    Y_i_obs={1}||Y[i,]; 								
    R_i_obs={1}||R[i,]; 							
    Y_i_miss=^R_i_obs	#(Y_i_obs*A);			
    Y_ij=Y_i_obs+Y_i_miss;					
    a_jk=diag(^R_i_obs)*A* diag(^R_i_obs); 
    T=T+Y_ij`*Y_ij+a_jk;		
  end;
  T=T/n;
  A=sweep(T,{1});
  A[,1]=-A[,1];
finish EMmap;

store;
quit;

/* Main Supplemented EM Macro */
%macro SEM(indata,var,nvar,nobs,calismodel);
  /* pre-process input data set */
  data misspt; set &indata; keep &var;
  run;
  data misspt; set misspt; 
    array Y(*) &var;
    array R(*) r1-r&nvar;
    do j=1 to &nvar;
      if Y(j) eq . then R(j)=0;
        else R(j)=1;		
    end;   		
    drop j;
  run;
  /* Use Proc MI to run EM algorithm */
  proc mi data=misspt nimpute=0 noprint;
    var &var;
    em maxiter=&maxEC outem=&emest;
  run;
  /* Fit the Calis model */
  %FitModel(nobs=&nobs,model=&calismodel);

  /**********************************************************/
  /*  collection of main subroutines used to compute Tb
  /*  compute EM estimates
  /*  compute &Delta& Matrix by Supplemented EM algorithm
  /*  compute observed complete data covariance matrix, V 
  /**********************************************************/
  proc iml;
    load;
    /***********************************************************/
    /*  construct initial input matrix for EM algorithm
    /*  theta_0-initial values	
    /*  his-iteration history is stored in it
    /*  last column contains mle estimated by EM algorithm
    /***********************************************************/
    use misspt;
    read all var {&var} into Y;
    read all var ('r1':"r&nvar") into R;
    mean=J(&nvar,1,0); cov=I(&nvar);
    theta_1={-1}||mean`;
    theta_2=mean||cov;
    theta=theta_1//theta_2;
    theta_0=theta;
    n=nrow(Y);
    p=ncol(Y);
    conv=0;
    it=0;
    run vech(v_cov, cov, &nvar);
    his= mean // v_cov;

    /***********************************************************/
    /*  EM ALGORITHM
    /*  A-output, theta_0-initial values 
    /*  thetaem-EM estimated MLE, augmented form
    /*  M-EM estimated mean vector
    /*  S-EM estimated covariance matrix
    /***********************************************************/
    do it=1 to &maxEC while(conv=0);
      run EMmap(A, theta_0, n, p, R, Y );		
      conv=all(abs(A-theta_0)< 1e-12);
      theta_0=A;
      em_mean=A[2:&nvar+1,1];
      em_cov=A[2:&nvar+1,2:&nvar+1];
      run vech(v_em_cov, em_cov, &nvar);
      his=his||(em_mean//v_em_cov);
    end;
    thetaem=A; 		
    M=A[2:&nvar+1,1];		
    S=A[2:&nvar+1,2:&nvar+1];
    print it [label="EM Iteration Number"];

    if conv<1 then 
      print, "****EM ITERATION LIMIT EXCEEDED. NOT CONVERGED.****";

    /**********************************************************/
    /*  Supplemented EM Algorithm
    /**********************************************************/
    pstar=&nvar*(&nvar+3)/2;
    /*  create indices for parameters 
    /*  with or without missing information 
    /*  varmis- indicator of missing information
    /*  pidxwomis-variable index without missing infomation
    /*  pidxwimis-variable index with missing information
    /*  d-number of variables with missing information */
    pidxwimis=j(pstar,1,0);
    varmis=R[+,] ^= &nobs;
    count=&nvar;
    do i=1 to &nvar;
      if varmis[i]=1 then pidxwimis[i]=1;
      do j=1 to i;
        count=count+1;
        if (varmis[i]=1) | (varmis[j]=1) then 
          pidxwimis[count]=1;
      end;
    end;	
    pidxwomis=loc(pidxwimis=0);
    pidxwimis=loc(pidxwimis=1);
    d=ncol(pidxwimis); 

    DM_0=j(d,d,0);
    DM=j(d,d,0);
    conv2=0;
    run r2v(th,thetaem,&nvar);

    /* forced-EM */
    do k=1 to it-1 while (conv2=0); *---(1);
      do ro=1 to d; *---(2);
        th[pidxwimis[ro],1]=His[pidxwimis[ro],k];	
        run v2r(rth,th,&nvar);
        run EMmap(Mj,rth,n,p,R,Y );
        run r2v(v_Mj,Mj,&nvar);
        a=v_Mj[pidxwimis]-HIS[pidxwimis,it]; 
        b=HIS[pidxwimis[ro],k]-HIS[pidxwimis[ro],it]; 
        c=a/b;
        DM[ro,]=c`;
        th=HIS[,it];
      end;
      conv2=all ( abs (DM-DM_0) < 1e-5);
      DM_0=DM;
    end;
    print k [label="SEM Iteration Number"]; 

    if conv2<1 then 
      print, "****SEM ITERATION LIMIT EXCEEDED. NOT CONVERGED.****";

    /**********************************************************/
    /*  complete data observed covariance matrix
    /**********************************************************/
    aa=inv(S);
    run Dup(Dp, &nvar);
    bb=1/2*Dp`*(aa@aa)*Dp;
    I_OC=block(aa,bb);
    COV_OC=inv(I_OC);	 

    /***********************************************************/
    /*  observed data covariance matrix, V
    /*	to be used in computing T_b
    /***********************************************************/
    G1=COV_OC[pidxwomis,pidxwomis];
    G2=COV_OC[pidxwomis,pidxwimis];
    G3=COV_OC[pidxwimis,pidxwimis];
    DV=(G3-G2`*inv(G1)*G2)*DM*inv(I(d)-DM);
    V=COV_OC;
    V[pidxwimis,pidxwimis]=G3+DV;
    
*	print V [label="Large Sample Covariance Matrix of MLE" format=10.4];
    create Vmatrix FROM V ;
    append from V;

    /**********************************************************/
    /*  compute asymptotically chi-squared GOF statistic, Tb
    /**********************************************************/
    use &emest;
    read all into y;
    yc=y[2:&nvar+1,1:&nvar];/*em estimated covariance matrix*/
    ym=y[1,]; /*em estimated mean vector*/		

    /*compute residuals*/
    use uncov;
    read all into uncov;/*uncorrected covariance analyzed*/
    use maxpred;
    read all into mlpred;/*ML predicted model value*/

    mlpred=mlpred[1:&nvar+1,];/*select the necessary submatrix*/			
    mlme=mlpred[&nvar+1,]`;/*ML predicted mean*/
    mlcov=mlpred[1:&nvar,1:&nvar]-mlme*mlme`;/*ML predicted covariance*/
    mlcov=mlcov//mlme`;
    /*convert ucov into corrected covariance mamtrix*/
    uncov=(uncov[1:&nvar,1:&nvar]-uncov[&nvar+1,]`*uncov[&nvar+1,])
      //uncov[&nvar+1,];
    rs=uncov-mlcov;/*residual*/ 

    /*corrected jacobian, O*/	
    use jacob;
    read all into bojac;
    p=nrow(bojac);
    q=ncol(bojac);
    J1=bojac[1:&nvar*(&nvar+1)/2, ];
    J2=bojac[&nvar*(&nvar+1)/2+1:&nvar*(&nvar+1)/2+&nvar,];
    O=J(&nvar*(&nvar+1)/2,q); /*corrected jacobian*/
    do k=1 to q;
      count=0;
      do i=1 to &nvar;
        do j=1 to i;
    	  count=count+1;
          correct= J2[i,k]*ym[1,j] + ym[1,i]*J2[j,k];
          O[count,k]=J1[count,k]-correct;
        end;
      end;
    end;

    O=J2//O;
    O=O#(abs(O)>1.0E-8);
    np=ncol(O);/*number of effective parameters*/ 
    ni=nrow(O);/*number of information given*/

    /*orthogonal complement, U s.t. U`*O=0*/
    P=I(pstar)-O* inv(O` * O) * O`; 
    call svd(u,q,h,P);
    df=ni-np;
    U=U[,1:df];

    /*residual covariance , rc */ 
    rc=rs[1:&nvar,1:&nvar]; 
    /*residual mean , rm */ 
    rm=rs[&nvar+1,];
    run vech(v_rc, rc, &nvar);
    v_rm=rm` // v_rc;

    /* asymptotically chi-squared GOF statistic */
    Tb=&nobs*v_rm`*U*inv(U`*V*U)*U`*v_rm;
    pv=1-probchi(Tb,df);
    Tblabels = {"T_B" "df" "p-value"};
    Tbresults = Tb||df||pv;
    print Tbresults [label="Goodness-of-fit Test for EM2S Estimator" 
      format=10.4 colname=Tblabels];
	
    /* compute point estiamte of RMSEA */
    Fhat=Tb/&nobs;
    RMSEA=max(sqrt(Fhat/df - inv(&nobs)),0);
    print RMSEA [label="Point Estimate of RMSEA" format=10.4];

    /* compute standard errors */
    acov = inv(&nobs*O`*inv(V)*O);
    se = sqrt(vecdiag(acov));

    use parmests;
    read all into parms [colname=parmnames];
    use wrongstderrs;
    read all into calisse;
    zvals = parms`/se;
    pvals = 2*(1-probnorm(abs(zvals)));
    columnlabels = {"Est." "Adj. SE" "Z" "Pr>|Z|" "Unadj. SE"};
    results = parms`||se||zvals||pvals||calisse`;
    print results [label="Adjusted Results for Structural Parameters" 
      format=10.3 colname=columnlabels rowname=parmnames];
  quit;
%mend;
